@@ -660,6 +660,10 @@ class MolterCommand:
             self._parameter_info,
         )
 
+        # we need to deal with subcommands too
+        for command in self.all_commands:
+            command._parse_parameters()
+
     def is_subcommand(self) -> bool:
         """Returns if this command is a subcommand or not."""
         return bool(self.parent)
